@@ -1,4 +1,4 @@
-#   Copyright (C) 2006-2012 Ludovic Rousseau (ludovic.rousseau@free.fr)
+#   Copyright (C) 2006-2015 Ludovic Rousseau (ludovic.rousseau@free.fr)
 #   Copyright (C) 2010 Giuseppe Amato (additions to original interface)
 #
 # This file is free software; you can redistribute it and/or modify it
@@ -13,9 +13,7 @@
 #
 # You should have received a copy of the GNU General Public License
 # along with this program; if not, write to the Free Software
-# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
-
-# $Id$
+# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA.
 
 from __future__ import print_function
 
@@ -27,10 +25,22 @@ PY3 = sys.version_info[0] >= 3
 if PY3:
     def byte_to_int(byte):
         return byte
+
+    def to_param_string(param):
+        if isinstance(param, str):
+            return bytes(param, 'ascii')
+        else:
+            return bytes(param)
 else:
     def byte_to_int(byte):
         return ord(byte)
 
+    def to_param_string(param):
+        if isinstance(param, str):
+            return param
+        else:
+            return str(bytearray(param))
+
     range = xrange
 
 # redefine PKCS#11 constants
@@ -86,20 +96,6 @@ class ckbytelist(PyKCS11.LowLevel.ckbytelist):
         return repr(rep)
 
 
-class byteArray(PyKCS11.LowLevel.byteArray):
-    """
-    add a __repr__() method to the LowLevel equivalent
-    """
-
-    def __repr__(self):
-        """
-        return the representation of a tuple
-        the __str__ method will use it also
-        """
-        rep = [elt for elt in self]
-        return repr(rep)
-
-
 class CK_OBJECT_HANDLE(PyKCS11.LowLevel.CK_OBJECT_HANDLE):
     """
     add a __repr__() method to the LowLevel equivalent
@@ -450,13 +446,18 @@ class PyKCS11Lib(object):
             pkcs11dll_filename = os.getenv("PYKCS11LIB")
             if pkcs11dll_filename == None:
                 raise PyKCS11Error(-1, "No PKCS11 library specified (set PYKCS11LIB env variable)")
-        rv = self.lib.Load(pkcs11dll_filename, 1)
+        rv = self.lib.Load(pkcs11dll_filename, True)
         if rv == 0:
             raise PyKCS11Error(-1, pkcs11dll_filename)
 
     def initToken(self, slot, pin, label):
         """
         C_InitToken
+
+        @param slot: slot number returned by L{getSlotList}
+        @type slot: integer
+        @param pin: SO's initial PIN
+        @param label: new label of the token
         """
         rv = self.lib.C_InitToken(slot, pin, label)
         if rv != CKR_OK:
@@ -578,7 +579,7 @@ class PyKCS11Lib(object):
 
         @param slot: slot number returned by L{getSlotList}
         @type slot: integer
-        @param flags: 0 (default), CKF_RW_SESSION for RW session
+        @param flags: 0 (default), L{CKF_RW_SESSION} for RW session
         @type flags: integer
         @return: a L{Session} object
         """
@@ -598,6 +599,8 @@ class PyKCS11Lib(object):
         """
         C_GetMechanismList
 
+        @param slot: slot number returned by L{getSlotList}
+        @type slot: integer
         @return: the list of available mechanisms for a slot
         @rtype: list
         """
@@ -608,13 +611,22 @@ class PyKCS11Lib(object):
 
         m = []
         for x in range(len(mechanismList)):
-            m.append(CKM[mechanismList[x]])
+            mechanism = mechanismList[x]
+            if mechanism >= CKM_VENDOR_DEFINED:
+                k = 'CKR_VENDOR_DEFINED_%X' % (mechanism - CKM_VENDOR_DEFINED)
+                CKM[k] = mechanism
+                CKM[mechanism] = k
+            m.append(CKM[mechanism])
         return m
 
     def getMechanismInfo(self, slot, type):
         """
         C_GetMechanismInfo
 
+        @param slot: slot number returned by L{getSlotList}
+        @type slot: integer
+        @param type: a CKM_* type
+        @type type: integer
         @return: information about a mechanism
         @rtype: a L{CK_MECHANISM_INFO} object
         """
@@ -663,6 +675,7 @@ class Mechanism(object):
         self.mechanism = mechanism
         self.param = param
 
+MechanismSHA1 = Mechanism(CKM_SHA_1, None)
 MechanismRSAPKCS1 = Mechanism(CKM_RSA_PKCS, None)
 MechanismRSAGENERATEKEYPAIR = Mechanism(CKM_RSA_PKCS_KEY_PAIR_GEN, None)
 
@@ -731,6 +744,8 @@ class Session(object):
     def initPin(self, new_pin):
         """
         C_InitPIN
+
+        @param new_pin: new PIN
         """
         rv = self.lib.C_InitPIN(self.session, new_pin)
         if rv != CKR_OK:
@@ -739,6 +754,9 @@ class Session(object):
     def setPin(self, old_pin, new_pin):
         """
         C_SetPIN
+
+        @param old_pin: old PIN
+        @param new_pin: new PIN
         """
         rv = self.lib.C_SetPIN(self.session, old_pin, new_pin)
         if rv != CKR_OK:
@@ -747,6 +765,8 @@ class Session(object):
     def createObject(self, template):
         """
         C_CreateObject
+
+        @param template: object template
         """
         attrs = self._template2ckattrlist(template)
         handle = PyKCS11.LowLevel.CK_OBJECT_HANDLE()
@@ -758,11 +778,60 @@ class Session(object):
     def destroyObject(self, obj):
         """
         C_DestroyObject
+
+        @param obj: object ID
         """
         rv = self.lib.C_DestroyObject(self.session, obj)
         if rv != CKR_OK:
             raise PyKCS11Error(rv)
 
+    def digest(self, data, mecha=MechanismSHA1):
+        """
+        C_DigestInit/C_Digest
+
+        @param data: the data to be digested
+        @type data:  (binary) sring or list/tuple of bytes
+        @param mecha: the digesting mechanism to be used
+        @type mecha: L{Mechanism} instance or L{MechanismSHA1}
+        for CKM_SHA_1
+        @return: the computed digest
+        @rtype: list of bytes
+
+        @note: the returned value is an istance of L{ckbytelist}.
+        You can easly convert it to a binary string with::
+            ''.join(chr(i) for i in ckbytelistDigest)
+
+        """
+        m = PyKCS11.LowLevel.CK_MECHANISM()
+        digest = ckbytelist()
+        ps = None  # must be declared here or may be deallocated too early
+        m.mechanism = mecha.mechanism
+        if (mecha.param):
+            # Convert the parameter to a string representation so SWIG gets a char*
+            ps = to_param_string(mecha.param)
+            m.pParameter = ps
+            m.ulParameterLen = len(mecha.param)
+        data1 = ckbytelist()
+        data1.reserve(len(data))
+        if isinstance(data, bytes):
+            for x in data:
+                data1.append(byte_to_int(x))
+        else:
+            for c in range(len(data)):
+                data1.append(data[c])
+        rv = self.lib.C_DigestInit(self.session, m)
+        if rv != CKR_OK:
+            raise PyKCS11Error(rv)
+        #first call get digest size
+        rv = self.lib.C_Digest(self.session, data1, digest)
+        if rv != CKR_OK:
+            raise PyKCS11Error(rv)
+        #second call get actual digest data
+        rv = self.lib.C_Digest(self.session, data1, digest)
+        if rv != CKR_OK:
+            raise PyKCS11Error(rv)
+        return digest
+
     def sign(self, key, data, mecha=MechanismRSAPKCS1):
         """
         C_SignInit/C_Sign
@@ -770,34 +839,26 @@ class Session(object):
         @param key: a key handle, obtained calling L{findObjects}.
         @type key: integer
         @param data: the data to be signed
-        @type data:  (binary) sring or list/tuple of bytes
+        @type data:  (binary) string or list/tuple of bytes
         @param mecha: the signing mechanism to be used
         @type mecha: L{Mechanism} instance or L{MechanismRSAPKCS1}
         for CKM_RSA_PKCS
         @return: the computed signature
         @rtype: list of bytes
 
-        @note: the returned value is an istance of L{ckbytelist}.
+        @note: the returned value is an instance of L{ckbytelist}.
         You can easly convert it to a binary string with::
             ''.join(chr(i) for i in ckbytelistSignature)
 
         """
         m = PyKCS11.LowLevel.CK_MECHANISM()
         signature = ckbytelist()
-        ba = None  # must be declared here or may be deallocated too early
+        ps = None  # must be declared here or may be deallocated too early
         m.mechanism = mecha.mechanism
         if (mecha.param):
-            ba = PyKCS11.LowLevel.byteArray(len(mecha.param))
-            if isinstance(mecha.param, bytes):
-                for c in range(len(mecha.param)):
-                    ba[c] = byte_to_int(mecha.param[c])
-            else:
-                for c in range(len(mecha.param)):
-                    ba[c] = mecha.param[c]
-            # with cast() the ba object continue to own internal pointer
-            # (avoids a leak).
-            # pParameter is an opaque pointer, never garbage collected.
-            m.pParameter = ba.cast()
+            # Convert the parameter to a string representation so SWIG gets a char*
+            ps = to_param_string(mecha.param)
+            m.pParameter = ps
             m.ulParameterLen = len(mecha.param)
         data1 = ckbytelist()
         data1.reserve(len(data))
@@ -820,6 +881,99 @@ class Session(object):
             raise PyKCS11Error(rv)
         return signature
 
+    def verify(self, key, data, signature, mecha=MechanismRSAPKCS1):
+        """
+        C_VerifyInit/C_Verify
+
+        @param key: a key handle, obtained calling L{findObjects}.
+        @type key: integer
+        @param data: the data that was signed
+        @type data:  (binary) string or list/tuple of bytes
+        @param signature: the signature to be verified
+        @type signature:  (binary) string or list/tuple of bytes
+        @param mecha: the signing mechanism to be used
+        @type mecha: L{Mechanism} instance or L{MechanismRSAPKCS1}
+        for CKM_RSA_PKCS
+        @return: True if signature is valid, False otherwise
+        @rtype: bool
+
+        """
+        m = PyKCS11.LowLevel.CK_MECHANISM()
+        ps = None  # must be declared here or may be deallocated too early
+        m.mechanism = mecha.mechanism
+        if (mecha.param):
+            # Convert the parameter to a string representation so SWIG gets a char*
+            ps = to_param_string(mecha.param)
+            m.pParameter = ps
+            m.ulParameterLen = len(mecha.param)
+        data1 = ckbytelist()
+        data1.reserve(len(data))
+        if isinstance(data, bytes):
+            for x in data:
+                data1.append(byte_to_int(x))
+        else:
+            for c in range(len(data)):
+                data1.append(data[c])
+        rv = self.lib.C_VerifyInit(self.session, m, key)
+        if rv != CKR_OK:
+            raise PyKCS11Error(rv)
+        rv = self.lib.C_Verify(self.session, data1, signature)
+        if rv == CKR_OK:
+            return True
+        elif rv == CKR_SIGNATURE_INVALID:
+            return False
+        else:
+            raise PyKCS11Error(rv)
+
+    def encrypt(self, key, data, mecha=MechanismRSAPKCS1):
+        """
+        C_EncryptInit/C_Encrypt
+
+        @param key: a key handle, obtained calling L{findObjects}.
+        @type key: integer
+        @param data: the data to be encrypted
+        @type data:  (binary) string or list/tuple of bytes
+        @param mecha: the encryption mechanism to be used
+        @type mecha: L{Mechanism} instance or L{MechanismRSAPKCS1}
+        for CKM_RSA_PKCS
+        @return: the encrypted data
+        @rtype: list of bytes
+
+        @note: the returned value is an instance of L{ckbytelist}.
+        You can easly convert it to a binary string with::
+            ''.join(chr(i) for i in ckbytelistEncrypted)
+
+        """
+        m = PyKCS11.LowLevel.CK_MECHANISM()
+        encrypted = ckbytelist()
+        ps = None  # must be declared here or may be deallocated too early
+        m.mechanism = mecha.mechanism
+        if (mecha.param):
+            # Convert the parameter to a string representation so SWIG gets a char*
+            ps = to_param_string(mecha.param)
+            m.pParameter = ps
+            m.ulParameterLen = len(mecha.param)
+        data1 = ckbytelist()
+        data1.reserve(len(data))
+        if isinstance(data, bytes):
+            for x in data:
+                data1.append(byte_to_int(x))
+        else:
+            for c in range(len(data)):
+                data1.append(data[c])
+        rv = self.lib.C_EncryptInit(self.session, m, key)
+        if rv != CKR_OK:
+            raise PyKCS11Error(rv)
+        #first call get encrypted size
+        rv = self.lib.C_Encrypt(self.session, data1, encrypted)
+        if rv != CKR_OK:
+            raise PyKCS11Error(rv)
+        #second call get actual encrypted data
+        rv = self.lib.C_Encrypt(self.session, data1, encrypted)
+        if rv != CKR_OK:
+            raise PyKCS11Error(rv)
+        return encrypted
+
     def decrypt(self, key, data, mecha=MechanismRSAPKCS1):
         """
         C_DecryptInit/C_Decrypt
@@ -827,34 +981,26 @@ class Session(object):
         @param key: a key handle, obtained calling L{findObjects}.
         @type key: integer
         @param data: the data to be decrypted
-        @type data:  (binary) sring or list/tuple of bytes
+        @type data:  (binary) string or list/tuple of bytes
         @param mecha: the decrypt mechanism to be used
         @type mecha: L{Mechanism} instance or L{MechanismRSAPKCS1}
         for CKM_RSA_PKCS
         @return: the decrypted data
         @rtype: list of bytes
 
-        @note: the returned value is an istance of L{ckbytelist}.
+        @note: the returned value is an instance of L{ckbytelist}.
         You can easly convert it to a binary string with::
             ''.join(chr(i) for i in ckbytelistData)
 
         """
         m = PyKCS11.LowLevel.CK_MECHANISM()
         decrypted = ckbytelist()
-        ba = None  # must be declared here or may be deallocated too early
+        ps = None  # must be declared here or may be deallocated too early
         m.mechanism = mecha.mechanism
         if (mecha.param):
-            ba = PyKCS11.LowLevel.byteArray(len(mecha.param))
-            if isinstance(mecha.param, bytes):
-                for c in range(len(mecha.param)):
-                    ba[c] = byte_to_int(mecha.param[c])
-            else:
-                for c in range(len(mecha.param)):
-                    ba[c] = mecha.param[c]
-            # with cast() the ba object continue to own internal pointer
-            # (avoids a leak).
-            # pParameter is an opaque pointer, never garbage collected.
-            m.pParameter = ba.cast()
+            # Convert the parameter to a string representation so SWIG gets a char*
+            ps = to_param_string(mecha.param)
+            m.pParameter = ps
             m.ulParameterLen = len(mecha.param)
         data1 = ckbytelist()
         data1.reserve(len(data))
@@ -877,6 +1023,84 @@ class Session(object):
             raise PyKCS11Error(rv)
         return decrypted
 
+    def wrapKey(self, wrappingKey, key, mecha=MechanismRSAPKCS1):
+        """
+        C_WrapKey
+
+        @param wrappingKey: a wrapping key handle
+        @type wrappingKey: integer
+        @param key: a handle of the key to be wrapped
+        @type key: integer
+        @param mecha: the encrypt mechanism to be used
+        @type mecha: L{Mechanism} instance or L{MechanismRSAPKCS1}
+        for CKM_RSA_PKCS
+        @return: the wrapped key bytes
+        @rtype: list of bytes
+
+        @note: the returned value is an instance of L{ckbytelist}.
+        You can easily convert it to a binary string with::
+            ''.join(chr(i) for i in ckbytelistData)
+
+        """
+        m = PyKCS11.LowLevel.CK_MECHANISM()
+        wrapped = ckbytelist()
+        ps = None  # must be declared here or may be deallocated too early
+        m.mechanism = mecha.mechanism
+        if (mecha.param):
+            # Convert the parameter to a string representation so SWIG gets a char*
+            ps = to_param_string(mecha.param)
+            m.pParameter = ps
+            m.ulParameterLen = len(mecha.param)
+        #first call get wrapped size
+        rv = self.lib.C_WrapKey(self.session, m, wrappingKey, key, wrapped)
+        if rv != CKR_OK:
+            raise PyKCS11Error(rv)
+        #second call get actual wrapped key data
+        rv = self.lib.C_WrapKey(self.session, m, wrappingKey, key, wrapped)
+        if rv != CKR_OK:
+            raise PyKCS11Error(rv)
+        return wrapped
+
+    def unwrapKey(self, unwrappingKey, wrappedKey, template, mecha=MechanismRSAPKCS1):
+        """
+        C_UnwrapKey
+
+        @param unwrappingKey: the unwrapping key handle
+        @type unwrappingKey: integer
+        @param wrappedKey: the bytes of the wrapped key
+        @type wrappedKey:  (binary) string or list/tuple of bytes
+        @param template: template for the unwrapped key
+        @param mecha: the decrypt mechanism to be used
+        @type mecha: L{Mechanism} instance or L{MechanismRSAPKCS1}
+        for CKM_RSA_PKCS
+        @return: the unwrapped key object
+        @rtype: integer
+
+        """
+        m = PyKCS11.LowLevel.CK_MECHANISM()
+        wrapped = ckbytelist()
+        ps = None  # must be declared here or may be deallocated too early
+        m.mechanism = mecha.mechanism
+        if (mecha.param):
+            # Convert the parameter to a string representation so SWIG gets a char*
+            ps = to_param_string(mecha.param)
+            m.pParameter = ps
+            m.ulParameterLen = len(mecha.param)
+        data1 = ckbytelist()
+        data1.reserve(len(wrappedKey))
+        if isinstance(wrappedKey, bytes):
+            for x in wrappedKey:
+                data1.append(byte_to_int(x))
+        else:
+            for c in range(len(wrappedKey)):
+                data1.append(wrappedKey[c])
+        handle = PyKCS11.LowLevel.CK_OBJECT_HANDLE()
+        attrs = self._template2ckattrlist(template)
+        rv = self.lib.C_UnwrapKey(self.session, m, unwrappingKey, data1, attrs, handle)
+        if rv != CKR_OK:
+            raise PyKCS11Error(rv)
+        return handle
+
     def isNum(self, type):
         """
         is the type a numerical value?
@@ -915,6 +1139,7 @@ class Session(object):
         """
         if type in (CKA_ALWAYS_SENSITIVE,
             CKA_DECRYPT,
+            CKA_DERIVE,
             CKA_ENCRYPT,
             CKA_EXTRACTABLE,
             CKA_HAS_RESET,
@@ -985,20 +1210,12 @@ class Session(object):
         ck_pub_handle = PyKCS11.LowLevel.CK_OBJECT_HANDLE()
         ck_prv_handle = PyKCS11.LowLevel.CK_OBJECT_HANDLE()
         m = PyKCS11.LowLevel.CK_MECHANISM()
-        ba = None # must be declared here or may be deallocated too early
+        ps = None # must be declared here or may be deallocated too early
         m.mechanism = mecha.mechanism
         if (mecha.param):
-            ba = PyKCS11.LowLevel.byteArray(len(mecha.param))
-            if isinstance(mecha.param, bytes):
-                for c in range(len(mecha.param)):
-                    ba[c] = byte_to_int(mecha.param[c])
-            else:
-                for c in range(len(mecha.param)):
-                    ba[c] = mecha.param[c]
-            # with cast() the ba object continue to own internal pointer
-            # (avoids a leak).
-            # pParameter is an opaque pointer, never garbage collected.
-            m.pParameter = ba.cast()
+            # Convert the parameter to a string representation so SWIG gets a char*
+            ps = to_param_string(mecha.param)
+            m.pParameter = ps
             m.ulParameterLen = len(mecha.param)
         rv = self.lib.C_GenerateKeyPair(self.session, m, tPub, tPriv,
             ck_pub_handle, ck_prv_handle)
@@ -1022,11 +1239,15 @@ class Session(object):
         # we search for 10 objects by default. speed/memory tradeoff
         result = PyKCS11.LowLevel.ckobjlist(10)
 
-        self.lib.C_FindObjectsInit(self.session, t)
+        rv = self.lib.C_FindObjectsInit(self.session, t)
+        if rv != CKR_OK:
+            raise PyKCS11Error(rv)
 
         res = []
         while True:
-            self.lib.C_FindObjects(self.session, result)
+            rv = self.lib.C_FindObjects(self.session, result)
+            if rv != CKR_OK:
+                raise PyKCS11Error(rv)
             for x in result:
                 # make a copy of the handle: the original value get
                 # corrupted (!!)
@@ -1036,7 +1257,9 @@ class Session(object):
             if len(result) == 0:
                 break
 
-        self.lib.C_FindObjectsFinal(self.session)
+        rv = self.lib.C_FindObjectsFinal(self.session)
+        if rv != CKR_OK:
+            raise PyKCS11Error(rv)
         return res
 
     def getAttributeValue(self, obj_id, attr, allAsBinary=False):
@@ -1162,7 +1385,7 @@ class Session(object):
         @param size: number of random bytes to get
         @type size: integer
 
-        @note: the returned value is an istance of L{ckbytelist}.
+        @note: the returned value is an instance of L{ckbytelist}.
         You can easly convert it to a binary string with::
             ''.join(chr(i) for i in random)
         """
