@@ -1,6 +1,6 @@
 #!/usr/bin/env python
 
-#   Copyright (C) 2006-2008 Ludovic Rousseau (ludovic.rousseau@free.fr)
+#   Copyright (C) 2006-2014 Ludovic Rousseau (ludovic.rousseau@free.fr)
 #
 # This file is free software; you can redistribute it and/or modify it
 # under the terms of the GNU General Public License as published by
@@ -14,7 +14,8 @@
 #
 # You should have received a copy of the GNU General Public License
 # along with this program; if not, write to the Free Software
-# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
+# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA.
+
 from __future__ import print_function
 
 import PyKCS11
@@ -39,7 +40,7 @@ def hexx(intval):
     return x
 
 
-def dump(src, length=8):
+def dump(src, length=16):
     FILTER = ''.join([(len(repr(chr(x))) == 3) and chr(x) or '.' for x in range(256)])
     N = 0
     result = ''
@@ -59,9 +60,10 @@ def usage():
     print("[-S][--sign]", end=' ')
     print("[-d][--decrypt]", end=' ')
     print("[-h][--help]", end=' ')
+    print()
 
 try:
-    opts, args = getopt.getopt(sys.argv[1:], "p:c:Sd:h", ["pin=", "lib=", "sign", "decrypt", "help"])
+    opts, args = getopt.getopt(sys.argv[1:], "p:c:Sdh", ["pin=", "lib=", "sign", "decrypt", "help"])
 except getopt.GetoptError:
     # print help information and exit:
     usage()
@@ -124,8 +126,9 @@ for s in slots:
         if pin_available:
             try:
                 session.login(pin=pin)
-            except:
-                print("login failed, exception:", str(sys.exc_info()[1]))
+            except PyKCS11.PyKCS11Error as e:
+                print("login failed, exception:", e)
+                break
 
         objects = session.findObjects()
         print()
@@ -143,37 +146,41 @@ for s in slots:
         # only use the integer values and not the strings like 'CKM_RSA_PKCS'
         all_attributes = [e for e in all_attributes if isinstance(e, int)]
 
+        n_obj = 1
         for o in objects:
             print()
-            print((red + "==================== Object: %d ====================" + normal) % o.value())
-            attributes = session.getAttributeValue(o, all_attributes)
+            print((red + "==================== Object: %d/%d (%d) ====================" + normal) % (n_obj, len(objects), o.value()))
+            n_obj += 1
+            try:
+                attributes = session.getAttributeValue(o, all_attributes)
+            except PyKCS11.PyKCS11Error as e:
+                continue
             attrDict = dict(list(zip(all_attributes, attributes)))
             if attrDict[PyKCS11.CKA_CLASS] == PyKCS11.CKO_PRIVATE_KEY \
                and attrDict[PyKCS11.CKA_KEY_TYPE] == PyKCS11.CKK_RSA:
                 m = attrDict[PyKCS11.CKA_MODULUS]
                 e = attrDict[PyKCS11.CKA_PUBLIC_EXPONENT]
                 if m and e:
-                    mx = eval(b'0x' + binascii.hexlify(''.join(chr(c) for c in m).encode('ascii')))
-                    ex = eval(b'0x' + binascii.hexlify(''.join(chr(c) for c in e).encode('ascii')))
+                    mx = eval(b'0x' + ''.join("%02X" %c for c in m))
+                    ex = eval(b'0x' + ''.join("%02X" %c for c in e))
                 if sign:
                     try:
                         toSign = b"12345678901234567890"  # 20 bytes, SHA1 digest
                         print("* Signing with object 0x%08X following data: %s" % (o.value(), toSign))
                         signature = session.sign(o, toSign)
-                        s = binascii.hexlify(''.join(chr(c) for c in signature).encode('ascii'))
-                        sx = eval(b'0x' + s)
+                        sx = eval(b'0x' + ''.join("%02X" % c for c in signature))
                         print("Signature:")
-                        print(dump(''.join(map(chr, signature)), 16))
+                        print(dump(''.join(map(chr, signature))))
                         if m and e:
                             print("Verifying using following public key:")
                             print("Modulus:")
-                            print(dump(''.join(map(chr, m)), 16))
+                            print(dump(''.join(map(chr, m))))
                             print("Exponent:")
-                            print(dump(''.join(map(chr, e)), 16))
+                            print(dump(''.join(map(chr, e))))
                             decrypted = pow(sx, ex, mx)  # RSA
                             print("Decrypted:")
                             d = binascii.unhexlify(hexx(decrypted))
-                            print(dump(d, 16))
+                            print(dump(d))
                             if toSign == d[-20:]:
                                 print("*** signature VERIFIED!\n")
                             else:
@@ -181,30 +188,35 @@ for s in slots:
                                 print(hex(decrypted), "\n")
                         else:
                             print("Unable to verify signature: MODULUS/PUBLIC_EXP not found")
-                    except:
-                        print("Sign failed, exception:", str(sys.exc_info()[1]))
+                    except PyKCS11.PyKCS11Error as e:
+                        print("Sign failed, exception:", e)
+                        break
                 if decrypt:
                     if m and e:
                         try:
                             toEncrypt = "12345678901234567890"
                             # note: PKCS1 BT2 padding should be random data,
                             # but this is just a test and we use 0xFF...
-                            padded = "\x00\x02%s\x00%s" % ("\xFF" * (128 - (len(toEncrypt)) - 3), toEncrypt)
-                            padded = padded.encode('latin-1')
+                            padded = "0002%s00%s" % ("FF" * (128 - (len(toEncrypt)) - 3), toEncrypt)
+                            padded = binascii.unhexlify(padded)
                             print("* Decrypting with 0x%08X following data: %s" % (o.value(), toEncrypt))
-                            print("padded:\n", dump(padded, 16))
+                            print("padded:")
+                            print(dump(padded))
                             encrypted = pow(eval('0x%sL' % binascii.hexlify(padded)), ex, mx)  # RSA
                             encrypted1 = binascii.unhexlify(hexx(encrypted))
-                            print("encrypted:\n", dump(encrypted1, 16))
+                            print("encrypted:")
+                            print(dump(encrypted1))
                             decrypted = session.decrypt(o, encrypted1)
                             decrypted1 = ''.join(chr(i) for i in decrypted)
-                            print("decrypted:\n", dump(decrypted1, 16))
+                            print("decrypted:")
+                            print(dump(decrypted1))
                             if decrypted1 == toEncrypt:
                                 print("decryption SUCCESSFULL!\n")
                             else:
                                 print("decryption FAILED!\n")
-                        except:
-                            print("Decrypt failed, exception:", str(sys.exc_info()[1]))
+                        except PyKCS11.PyKCS11Error as e:
+                            print("Decrypt failed, exception:", e)
+                            break
                     else:
                         print("ERROR: Private key don't have MODULUS/PUBLIC_EXP")
 
@@ -222,11 +234,11 @@ for s in slots:
                 elif session.isBin(q):
                     print(format_binary % (PyKCS11.CKA[q], len(a)))
                     if a:
-                        print(dump(''.join(map(chr, a)), 16), end=' ')
+                        print(dump(''.join(map(chr, a))), end='')
                 elif q == PyKCS11.CKA_SERIAL_NUMBER:
                     print(format_binary % (PyKCS11.CKA[q], len(a)))
                     if a:
-                        print(dump(a, 16), end=' ')
+                        print(dump(a), end='')
                 else:
                     print(format_normal % (PyKCS11.CKA[q], a))
         print()
@@ -234,8 +246,9 @@ for s in slots:
         if pin_available:
             try:
                 session.logout()
-            except:
-                print("logout failed, exception:", str(sys.exc_info()[1]))
+            except PyKCS11.PyKCS11Error as e:
+                print("logout failed, exception:", e)
+                break
 
         session.closeSession()
 
